// Copyright 2002-2003 University of Guelph, All Rights Reserved
import java.util.*;

public class Problem {
  public static final int MAX_ZEBRA = 5;
  public static final int MAX_SHERLOCK = 6;
  public static final int MAX_ZEBRA_VARS = 25;
  public static final int MAX_SHERLOCK_VARS = 36;

  // Problem type
  public static final int ZEBRA = 0;
  public static final int SHERLOCK = 1;

  // Relation type
  public static final int SAME_COL = 0;
  public static final int NOT_SAME_COL = 1;
  public static final int NEXT_TO = 2;
  public static final int NOT_NEXT_TO = 3;
  public static final int LEFT_OF = 4;
  public static final int NOT_EQUAL = 7;
  public static final int NIL = 8;
  public static final int RIGHT_OF = 9;
  public static final int NOT_NEXT_TO_SAME = 10;
  public static final int GIVEN = 5;
  public static final int NEXT_RIGHT = 6;
  public static final int NEXT_LEFT = 11;

  // Solution value
  public static final int NO_SOLUTION = -1;

  long numNodes;
  HashSet constraints[][];
  int solution[];
  HashSet domain[];
  int problemType;
  long numChecks;

  public int getType() {
	return problemType;
  }

  public int getMaxDomain() {
	if (problemType == ZEBRA) {
	  return MAX_ZEBRA;
	} else {
	  return MAX_SHERLOCK;
	}
  }

  public int getMaxVars() {
	if (problemType == ZEBRA) {
	  return MAX_ZEBRA_VARS;
	} else {
	  return MAX_SHERLOCK_VARS;
	}
  }

  Problem(int Type) {
	problemType = Type;

	int maxVars = getMaxVars();

	numChecks = 0;
	numNodes = 0;
	constraints = new HashSet[maxVars + 1][maxVars + 1];
	
	for (int i = 0; i <= maxVars; i++) {
	  for (int j = 0; j <= maxVars; j++) {
		constraints[i][j] = null;
	  }
	}


	solution = new int[maxVars + 1];
	domain = new HashSet[maxVars + 1];
	domain[0] = null;

	for (int i = 1; i <= maxVars; i++) {
	  solution[i] = NO_SOLUTION;

	  int maxDomain = getMaxDomain();

	  domain[i] = new HashSet();

	  for (int j = 1; j <= maxDomain; j++) {
		domain[i].add(new Integer(j));
	  }
	}
  }
  
  public void givenVariable(int VariableNum, int Value) {
	domain[VariableNum].clear();
	domain[VariableNum].add(new Integer(Value));
  }

  public void newRelation(int V1, String RelationString, int V2) throws Exception {
	int relationType;
	
	checkVariable(V1);
	checkVariable(V2);

	relationType = NIL;

	if (RelationString.equals("not-equal")) {
	  relationType = NOT_EQUAL;
	} else if (RelationString.equals("same-col")) {
	  relationType = SAME_COL;
	} else if (RelationString.equals("next-to")) {
	  relationType = NEXT_TO;
	} else if (RelationString.equals("next-right")) {
	  relationType = NEXT_RIGHT;
	} else if (RelationString.equals("next-left")) {
	  relationType = NEXT_LEFT;
	} else if (problemType == SHERLOCK) {
	  if (RelationString.equals("not-same-col")) {
		relationType = NOT_SAME_COL;
	  } else if (RelationString.equals("not-next-to")) {
		relationType = NOT_NEXT_TO;
	  } else if (RelationString.equals("left-of")) {
		relationType = LEFT_OF;
	  } else if (RelationString.equals("right-of")) {
		relationType = RIGHT_OF;
	  } else if (RelationString.equals("not-next-same")) {
		relationType = NOT_NEXT_TO_SAME; 
	  } else {
		if (Debug.DEBUG_LEVEL >= Debug.DEBUG_VERBOSE) {
		  System.out.println("Not a valid relation.");
		}
		throw (new Exception("Not a valid relation for a Zebra problem."));
	  }
	} else {
	  if (Debug.DEBUG_LEVEL >= Debug.DEBUG_VERBOSE) {
		System.out.println("Uknown relation or only two vars for between/not-between");
	  }
	  throw (new Exception("Unknown relation type or between/not-between missing third variable"));
	}
	addConstraint(V1, V2, relationType);
  }

  public void addConstraint(int V1, int V2, int Constraint) {
	if (constraints[V1][V2] == null) {
	  constraints[V1][V2] = new HashSet();
	}
	constraints[V1][V2].add(new Integer(Constraint));
  }

  public void checkVariable(int variable) throws Exception {
	if (variable < 1) {
	  throw (new Exception("Invalid variable value: must be 1 or greater."));
	}
	if (variable > getMaxVars()) {
	  throw (new Exception("Invalid variable value: must be less than " +
						  String.valueOf(getMaxVars())));
	}
  }
  public void newRelation(int V1, String RelationString, int V2, int V3) throws Exception {
	checkVariable(V1);
	checkVariable(V2);
	checkVariable(V3);

	if (RelationString.equals("between")) {
	  addConstraint(V2, V1, NEXT_TO);
	  addConstraint(V1, V2, NEXT_TO);
	  addConstraint(V1, V3, NEXT_TO);
	  addConstraint(V3, V1, NEXT_TO);
	  addConstraint(V2, V3, NOT_SAME_COL);
	  addConstraint(V3, V2, NOT_SAME_COL);
	} else {
	  throw new Exception("Unknown relation type.");
	}
  }
  
  boolean checkSolution() {
	return (checkConflicts() == 0);
  }

  public static Object getFromSet(Set set, int index) {
	int i = 0;
	Iterator itemIterator = set.iterator();
	Object item = null;

	while (itemIterator.hasNext() && i <= index) {
	  item = itemIterator.next();
	  i++;
	}
	if (Debug.DEBUG_LEVEL >= Debug.DEBUG_VERBOSE) {
	  System.out.println("Item " + item + " i: " + String.valueOf(i) + 
						 " index: " + String.valueOf(index));
	}
	
	return item;
  }

  public int checkConflicts() {
	int i, j;
	int maxI, maxJ;
	int numConflicts = 0;

	maxI = getMaxVars();
	maxJ = getMaxVars();

	for (i = 1; i <= maxI; i++) {
	  for (j = 1; j <= maxJ; j++) {
		if (i != j) {
		  if (!(check(i, j))) {
			numConflicts++;
		  }
		}
	  }
	}

	if (numConflicts > 1) {
	  numConflicts = numConflicts / 2;
	}
	for (i = 1; i <= maxI; i++) {
	  if (!(domain[i].contains(new Integer(solution[i])))) {
		if (Debug.DEBUG_LEVEL >= Debug.DEBUG_VERBOSE) {
		  System.out.println("Discarding solution because domain[" +
							 String.valueOf(i) + "] does not contain " +
							 String.valueOf(solution[i]));			
		}
		numConflicts++;
	  }		
	}

	return numConflicts;
  }

  public int varRow(int Variable) {
	int numRows = getMaxDomain();
	return ((Variable - 1) / numRows);
  }

  public HashSet getRowSet(int rowNum) {
	HashSet newSet = new HashSet();
	int numRows = getMaxDomain();;

	for (int i = 0; i < numRows; i++) {
	  newSet.add(new Integer(rowNum * numRows + i + 1));
	}
	return newSet;
  }

  public boolean checkConsistent(int Vi, int Vj, int I, int J) {
	boolean result = true;

	if (constraints[I][J] == null) {
	  result = true;
	} else {
	  numChecks++;

	  Iterator iterator = constraints[I][J].iterator();

	  while (iterator.hasNext()) {

		switch(((Integer)(iterator.next())).intValue()) {
		  case SAME_COL:
			if (Vi != Vj) {
			  result = false;
			} 
			break;
		  case NOT_SAME_COL:
			if (Vi == Vj) {
			result = false;
			}
			break;
		  case NEXT_TO:
			if (((Vi - Vj) !=  1) &&
				((Vi - Vj) != -1)) {
			  result = false;
			}
			break;	
			
		  case NOT_NEXT_TO:
			if (((Vi - Vj) ==  1) ||
				((Vi - Vj) == -1)) {
			  result = false;
			}
			break;
		  case LEFT_OF:
			if (Vi >= Vj) {
			  result = false;
			}
			break;
		  case RIGHT_OF:
			if (Vi <= Vj) {
			  result = false;
			}
			break;
		  case NOT_EQUAL:
			if (Vi == Vj) {
			  result = false;
			}
			break;
		  case NOT_NEXT_TO_SAME:
			if (((Vi - Vj) ==  1) ||
				((Vi - Vj) == -1) ||
				((Vi == Vj)))
			  {
				result = false;
			  }
			break;
		  case NEXT_RIGHT:
			if ((Vi - Vj) !=  1) { 
			  result = false;
			}
			break;
		  case NEXT_LEFT:
			if ((Vi - Vj) !=  -1) {
			  result = false;
			}
			break;
		}
	  }
	}
// 	if (Debug.DEBUG_LEVEL >= Debug.DEBUG_VERBOSE) {
// 	  if (!result) {
// 		System.out.print("i = " + String.valueOf(I) + " , j = " + 
// 						 String.valueOf(J) + " failed on ");
// 		switch(constraints[I][J]) {
// 		  case SAME_COL:
// 			System.out.println("same-col");
// 			break;
// 		  case NOT_SAME_COL:
// 			System.out.println("not-same-col");
// 			break;
// 		  case NEXT_TO:
// 			System.out.println("next-to");
// 			break;	
// 		  case NOT_NEXT_TO:
// 			System.out.println("not-next-to");
// 			break;
// 		  case LEFT_OF:
// 			System.out.println("left-of");
// 			break;
// 		  case RIGHT_OF:
// 			System.out.println("right-of");
// 			break;
// 		  case NOT_EQUAL:
// 			System.out.println("not-equal");
// 			break;
// 		  case NOT_NEXT_TO_SAME:
// 			System.out.println("not-next-same");
// 		  case NEXT_RIGHT:
// 			System.out.println("next-right");
// 		  case NEXT_LEFT:
// 			System.out.println("next-left");
// 		}
//	  }
//	}
	return result;
  }

  public boolean check(int i, int j) {
	return checkConsistent(solution[i], solution[j], i, j);
  }

  public String getSolutionString() {
	Hashtable colToLetter = new Hashtable();
	String colString;
	int i = 1;
	String outString = "";

	if (problemType == ZEBRA) {
	  colString = String.valueOf(solution[i++]);
	  colToLetter.put(colString, "R");
	  colString = String.valueOf(solution[i++]);
	  colToLetter.put(colString, "B");
	  colString = String.valueOf(solution[i++]);
	  colToLetter.put(colString, "Y");
	  colString = String.valueOf(solution[i++]);
	  colToLetter.put(colString, "G");
	  colString = String.valueOf(solution[i++]);
	  colToLetter.put(colString, "I");
	  outString += getRowString(colToLetter);
	  colString = String.valueOf(solution[i++]);
	  colToLetter.put(colString, "O");
	  colString = String.valueOf(solution[i++]);
	  colToLetter.put(colString, "P");
	  colString = String.valueOf(solution[i++]);
	  colToLetter.put(colString, "K");
	  colString = String.valueOf(solution[i++]);
	  colToLetter.put(colString, "L");
	  colString = String.valueOf(solution[i++]);
	  colToLetter.put(colString, "C");
	  outString += getRowString(colToLetter);
	  colString = String.valueOf(solution[i++]);
	  colToLetter.put(colString, "N");
	  colString = String.valueOf(solution[i++]);
	  colToLetter.put(colString, "U");
	  colString = String.valueOf(solution[i++]);
	  colToLetter.put(colString, "E");
	  colString = String.valueOf(solution[i++]);
	  colToLetter.put(colString, "S");
	  colString = String.valueOf(solution[i++]);
	  colToLetter.put(colString, "J");
	  outString += getRowString(colToLetter);
	  colString = String.valueOf(solution[i++]);
	  colToLetter.put(colString, "Z");
	  colString = String.valueOf(solution[i++]);
	  colToLetter.put(colString, "D");
	  colString = String.valueOf(solution[i++]);
	  colToLetter.put(colString, "H");
	  colString = String.valueOf(solution[i++]);
	  colToLetter.put(colString, "F");
	  colString = String.valueOf(solution[i++]);
	  colToLetter.put(colString, "S");
	  outString += getRowString(colToLetter);
	  colString = String.valueOf(solution[i++]);
	  colToLetter.put(colString, "C");
	  colString = String.valueOf(solution[i++]);
	  colToLetter.put(colString, "T");
	  colString = String.valueOf(solution[i++]);
	  colToLetter.put(colString, "W");
	  colString = String.valueOf(solution[i++]);
	  colToLetter.put(colString, "M");
	  colString = String.valueOf(solution[i++]);
	  colToLetter.put(colString, "O");
	  outString += getRowString(colToLetter);
	} else {
	  colString = String.valueOf(solution[i++]);
	  colToLetter.put(colString, "R");
	  colString = String.valueOf(solution[i++]);
	  colToLetter.put(colString, "B");
	  colString = String.valueOf(solution[i++]);
	  colToLetter.put(colString, "Y");
	  colString = String.valueOf(solution[i++]);
	  colToLetter.put(colString, "G");
	  colString = String.valueOf(solution[i++]);
	  colToLetter.put(colString, "W");
	  colString = String.valueOf(solution[i++]);
	  colToLetter.put(colString, "N");
	  outString += getRowString(colToLetter);
	  colString = String.valueOf(solution[i++]);
	  colToLetter.put(colString, "N");
	  colString = String.valueOf(solution[i++]);
	  colToLetter.put(colString, "U");
	  colString = String.valueOf(solution[i++]);
	  colToLetter.put(colString, "E");
	  colString = String.valueOf(solution[i++]);
	  colToLetter.put(colString, "S");
	  colString = String.valueOf(solution[i++]);
	  colToLetter.put(colString, "J");
	  colString = String.valueOf(solution[i++]);
	  colToLetter.put(colString, "A");
	  outString += getRowString(colToLetter);
	  colString = String.valueOf(solution[i++]);
	  colToLetter.put(colString, "1");
	  colString = String.valueOf(solution[i++]);
	  colToLetter.put(colString, "2");
	  colString = String.valueOf(solution[i++]);
	  colToLetter.put(colString, "3");
	  colString = String.valueOf(solution[i++]);
	  colToLetter.put(colString, "4");
	  colString = String.valueOf(solution[i++]);
	  colToLetter.put(colString, "5");
	  colString = String.valueOf(solution[i++]);
	  colToLetter.put(colString, "6");
	  outString += getRowString(colToLetter);
	  colString = String.valueOf(solution[i++]);
	  colToLetter.put(colString, "A");
	  colString = String.valueOf(solution[i++]);
	  colToLetter.put(colString, "O");
	  colString = String.valueOf(solution[i++]);
	  colToLetter.put(colString, "B");
	  colString = String.valueOf(solution[i++]);
	  colToLetter.put(colString, "P");
	  colString = String.valueOf(solution[i++]);
	  colToLetter.put(colString, "S");
	  colString = String.valueOf(solution[i++]);
	  colToLetter.put(colString, "C");
	  outString += getRowString(colToLetter);
	  colString = String.valueOf(solution[i++]);
	  colToLetter.put(colString, "S");
	  colString = String.valueOf(solution[i++]);
	  colToLetter.put(colString, "R");
	  colString = String.valueOf(solution[i++]);
	  colToLetter.put(colString, "O");
	  colString = String.valueOf(solution[i++]);
	  colToLetter.put(colString, "P");
	  colString = String.valueOf(solution[i++]);
	  colToLetter.put(colString, "H");
	  colString = String.valueOf(solution[i++]);
	  colToLetter.put(colString, "D");
	  outString += getRowString(colToLetter);
	  colString = String.valueOf(solution[i++]);
	  colToLetter.put(colString, "H");
	  colString = String.valueOf(solution[i++]);
	  colToLetter.put(colString, "O");
	  colString = String.valueOf(solution[i++]);
	  colToLetter.put(colString, "L");
	  colString = String.valueOf(solution[i++]);
	  colToLetter.put(colString, "M");
	  colString = String.valueOf(solution[i++]);
	  colToLetter.put(colString, "E");
	  colString = String.valueOf(solution[i++]);
	  colToLetter.put(colString, "S");
	  outString += getRowString(colToLetter);
	}
	return outString;
  }

  public String getRowString(Hashtable colToLetter) {
	String outString = "";
	outString += ((String)(colToLetter.get(new String("1"))) + " ");	
	outString += ((String)(colToLetter.get(new String("2"))) + " ");	
	outString += ((String)(colToLetter.get(new String("3"))) + " ");	
	outString += ((String)(colToLetter.get(new String("4"))) + " ");	
	outString += ((String)(colToLetter.get(new String("5"))) + " ");
	if (problemType == SHERLOCK) {
	  outString += ((String)(colToLetter.get(new String("6"))) + " ");
	}
	outString += "\n";
	return outString;
  }

  public void arc3() {
	class arc {
	  int l, m;
	  
	  arc(int L, int M) {
		l = L;
		m = M;
	  }
	}

	HashSet Q = new HashSet();

	// Get non-nil constraints (i != j) and put in set Q
	for (int i = 1; i <= getMaxVars(); i++) {
	  for (int j = 1; j <= getMaxVars(); j++) {
		if (i != j) {
		  if (constraints[i][j] != null) {
			Q.add(new arc(i, j));
		  }
		}
	  }
	}
	
	HashSet G = (HashSet)(Q.clone());

	while (!Q.isEmpty()) {
	  int k, m;
	  arc ai;

	  Iterator qi = Q.iterator();
	  ai = (arc)(qi.next());
	  k = ai.l;
	  m = ai.m;

	  Q.remove(ai);
	  
	  if (revise(k, m)) {
		HashSet newQ = new HashSet();
		Iterator newIterator = G.iterator();

		while (newIterator.hasNext()) {
		  arc ag = (arc)(newIterator.next());

		  if ((ag.l != k) && (ag.l != m)) {
			Q.add(ag);
		  }
		}
	  }
	}
  }

  public boolean revise(int I, int J) {
	int x;

	boolean delete = false;
	HashSet toDelete = new HashSet();

	Iterator di = domain[I].iterator();

	while (di.hasNext()) {
	  boolean isConsistent = false;
	  int vj;

	  x = ((Integer)(di.next())).intValue();

	  Iterator dj = domain[J].iterator();
	  
	  while (dj.hasNext()) {
		vj = ((Integer)(dj.next())).intValue();

		if (checkConsistent(x, vj, I, J)) {
		  isConsistent = true;
		}
	  }

	  if (!isConsistent) {
		toDelete.add(new Integer(x));
		delete = true;
	  }
	}

	if (delete) {
	  Iterator delit = toDelete.iterator();

	  while (delit.hasNext()) {
		domain[I].remove(delit.next());
	  }
	}
	return delete;
  }

  public String getNodeString(int I, int V) {
	String nodeString = "";
	int tempSolution[] = (int [])(solution.clone());
	tempSolution[I] = V;
	for (int j = 0; j < tempSolution.length; j++) {
	  if (tempSolution[j] != 0) {
		nodeString += String.valueOf(tempSolution[j]);
	  }
	}
	return nodeString;
  }

  public void printSolution() {
	System.out.println(getSolutionString());
  }

  public String getConstraintString(int v1, int v2) {
	String out = "";

	Iterator iterator = constraints[v1][v2].iterator();

	while (iterator.hasNext()) {
	  out += String.valueOf(v1) + " ";

	  switch(((Integer)(iterator.next())).intValue()) {
		case SAME_COL:
		  out += "same-col ";
		  break;
		case NOT_SAME_COL:
		  out += "not-same-col ";
		  break;
		case NEXT_TO:
		  out += "next-to ";
		  break;
		case NOT_NEXT_TO:
		  out += "not-next-to ";
		  break;
		case LEFT_OF:
		  out += "left-of ";
		  break;
		case NOT_EQUAL:
		  out += "not-equal ";
		  break;
		case NIL:
		  out += "nil as clue type ";
		case RIGHT_OF:
		  out += "right-of ";
		  break;
		case NOT_NEXT_TO_SAME:
		  out += "not-next-same ";
		  break;
		case NEXT_RIGHT:
		  out += "next-right ";
		  break;
		case NEXT_LEFT:
		  out += "next-left ";
		  break;
	  }
	  out += String.valueOf(v2);
	  out += "\n";
	}
	return out;
  }
}
